# -*- coding: utf-8 -*-
"""
@author: Yuxian Meng
@contact: yuxian_meng@shannonai.com

@version: 1.0
@file: glyph_embedding
@time: 2020/8/4 15:04

"""


from torch import nn


class GlyphEmbedding(nn.Module):
    """Glyph2Image Embedding"""

    def __init__(self, num_embeddings, embedding_dim):
        super(GlyphEmbedding, self).__init__()
        self.embedding = nn.Embedding(
            num_embeddings=num_embeddings, embedding_dim=embedding_dim
        )

    def forward(self, input_ids):
        """
            get glyph images for batch inputs
        Args:
            input_ids: [batch, sentence_length]
        Returns:
            images: [batch, sentence_length, self.font_num*self.font_size*self.font_size]
        """
        # return self.embedding(input_ids).reshape([-1, self.font_num, self.font_size, self.font_size])
        return self.embedding(input_ids)



import json
import random
from datetime import datetime
from pathlib import Path
from typing import Dict, List, Optional, Union

import numpy as np
import pandas as pd
import torch
from fastcore.all import *
from transformers.file_utils import (
    PaddingStrategy,
    TensorType,
    _is_tensorflow,
    _is_torch,
    _is_torch_device,
    is_tf_available,
    is_torch_available,
    to_py_obj,
    torch_required,
)
from transformers.tokenization_utils_base import (
    BatchEncoding,
    EncodedInput,
    PreTrainedTokenizerBase,
)
from transformers.utils.logging import get_logger

logger = get_logger(__name__)


@patch_to(BatchEncoding)
@torch_required
def to(self, device: Union[str, "torch.device"]) -> "BatchEncoding":
    """
    Send all values to device by calling :obj:`v.to(device)` (PyTorch only).

    Args:
        device (:obj:`str` or :obj:`torch.device`): The device to put the tensors on.

    Returns:
        :class:`~transformers.BatchEncoding`: The same instance after modification.
    """

    # This check catches things like APEX blindly calling "to" on all inputs to a module
    # Otherwise it passes the casts down and casts the LongTensor containing the token idxs
    # into a HalfTensor
    if isinstance(device, str) or _is_torch_device(device) or isinstance(device, int):
        self.data = {
            k: [{subk: subv.to(device=device) for subk, subv in vv.items()} for vv in v]
            if isinstance(v, (tuple, list, L)) and isinstance(v[0], dict)
            else v.to(device=device)
            for k, v in self.data.items()
        }
    else:
        logger.warning(
            f"Attempting to cast a BatchEncoding to type {str(device)}. This is not supported."
        )
    return self


@patch_to(PreTrainedTokenizerBase)
def pad(
    self,
    encoded_inputs: Union[
        BatchEncoding,
        List[BatchEncoding],
        Dict[str, EncodedInput],
        Dict[str, List[EncodedInput]],
        List[Dict[str, EncodedInput]],
    ],
    padding: Union[bool, str, PaddingStrategy] = True,
    max_length: Optional[int] = None,
    pad_to_multiple_of: Optional[int] = None,
    return_attention_mask: Optional[bool] = None,
    return_tensors: Optional[Union[str, TensorType]] = None,
    verbose: bool = True,
    model_input_names: List = None,
):
    if model_input_names is not None:
        copy_model_input_names = self.model_input_names
        self.model_input_names = (
            [model_input_names]
            if isinstance(model_input_names, str)
            else model_input_names
        )

    if isinstance(encoded_inputs, (list, tuple)) and isinstance(
        encoded_inputs[0], (dict, BatchEncoding)
    ):
        encoded_inputs = {
            key: [example[key] for example in encoded_inputs]
            for key in encoded_inputs[0].keys()
            if key in self.model_input_names
        }

    # The model's main input name, usually `input_ids`, has be passed for padding
    if self.model_input_names[0] not in encoded_inputs:
        raise ValueError(
            "You should supply an encoding or a list of encodings to this method "
            f"that includes {self.model_input_names[0]}, but you provided {list(encoded_inputs.keys())}"
        )

    required_input = encoded_inputs[self.model_input_names[0]]

    if not required_input:
        if return_attention_mask:
            encoded_inputs["attention_mask"] = []
        return encoded_inputs

    # If we have PyTorch/TF/NumPy tensors/arrays as inputs, we cast them as python objects
    # and rebuild them afterwards if no return_tensors is specified
    # Note that we lose the specific device the tensor may be on for PyTorch

    first_element = required_input[0]
    if isinstance(first_element, (list, tuple)):
        # first_element might be an empty list/tuple in some edge cases so we grab the first non empty element.
        index = 0
        while len(required_input[index]) == 0:
            index += 1
        if index < len(required_input):
            first_element = required_input[index][0]
    # At this state, if `first_element` is still a list/tuple, it's an empty one so there is nothing to do.
    if not isinstance(first_element, (int, list, tuple)):
        if is_tf_available() and _is_tensorflow(first_element):
            return_tensors = "tf" if return_tensors is None else return_tensors
        elif is_torch_available() and _is_torch(first_element):
            return_tensors = "pt" if return_tensors is None else return_tensors
        elif isinstance(first_element, np.ndarray):
            return_tensors = "np" if return_tensors is None else return_tensors
        else:
            raise ValueError(
                f"type of {first_element} unknown: {type(first_element)}. "
                f"Should be one of a python, numpy, pytorch or tensorflow object."
            )

        for key, value in encoded_inputs.items():
            encoded_inputs[key] = to_py_obj(value)

    # Convert padding_strategy in PaddingStrategy
    padding_strategy, _, max_length, _ = self._get_padding_truncation_strategies(
        padding=padding, max_length=max_length, verbose=verbose
    )

    required_input = encoded_inputs[self.model_input_names[0]]
    if required_input and not isinstance(required_input[0], (list, tuple)):
        encoded_inputs = self._pad(
            encoded_inputs,
            max_length=max_length,
            padding_strategy=padding_strategy,
            pad_to_multiple_of=pad_to_multiple_of,
            return_attention_mask=return_attention_mask,
        )
        return BatchEncoding(encoded_inputs, tensor_type=return_tensors)

    batch_size = len(required_input)
    assert all(
        len(v) == batch_size for v in encoded_inputs.values()
    ), "Some items in the output dictionary have a different batch size than others."

    if padding_strategy == PaddingStrategy.LONGEST:
        max_length = max(len(inputs) for inputs in required_input)
        padding_strategy = PaddingStrategy.MAX_LENGTH

    batch_outputs = {}
    for i in range(batch_size):
        inputs = dict((k, v[i]) for k, v in encoded_inputs.items())
        outputs = self._pad(
            inputs,
            max_length=max_length,
            padding_strategy=padding_strategy,
            pad_to_multiple_of=pad_to_multiple_of,
            return_attention_mask=return_attention_mask,
        )

        for key, value in outputs.items():
            if key not in batch_outputs:
                batch_outputs[key] = []
            batch_outputs[key].append(value)

    self.model_input_names = copy_model_input_names
    return BatchEncoding(batch_outputs, tensor_type=return_tensors)


def get_current_time():
    return datetime.now().strftime("%Y-%m-%d-%H-%M-%S")


# def get_logger():
#     log = logging.getLogger(__name__)
#     logging.basicConfig(
#         format="%(asctime)s - %(levelname)s - %(name)s - %(message)s",
#         datefmt="%m/%d/%Y %H:%M:%S",
#         level=logging.INFO,
#         handlers=[
#             logging.FileHandler(
#                 "{}.log".format(get_current_time()),
#                 mode="w",
#                 encoding="utf-8",
#             )
#         ],
#     )
#     return log


def mk_dir(path, parents=True, exist_ok=True):
    p = Path(path)
    p.mkdir(parents=parents, exist_ok=exist_ok)
    return p


def seed_everything(seed):
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    return seed


def load_json(fn):
    "Load a json file from a file name or opened file"
    with open_file(fn, "r", encoding="utf8") as f:
        return json.load(f)


def save_json(fn, o):
    "Save a json file, to a file name or opened file"
    with open_file(fn, "w", encoding="utf8") as f:
        return json.dump(o, f, ensure_ascii=False)


def load_csv(fn, **kwargs):
    "Load a csv file from a file name"
    kwargs.pop("sep", None)
    return pd.read_csv(fn, **kwargs)


def load_tsv(fn, **kwargs):
    "Load a tsv file from a file name"
    kwargs.pop("sep", None)
    return pd.read_csv(fn, sep="\t", **kwargs)



from transformers import PretrainedConfig


class ChineseBertConfig(PretrainedConfig):
    r"""
    Args:
        vocab_size (:obj:`int`, `optional`, defaults to 23236):
            Vocabulary size of the BERT model. Defines the number of different tokens that can be represented by the
            :obj:`inputs_ids` passed when calling :class:`~ChineseBertModel` or
            :class:`~TFChineseBertModel`.
        hidden_size (:obj:`int`, `optional`, defaults to 768):
            Dimensionality of the encoder layers and the pooler layer.
        num_hidden_layers (:obj:`int`, `optional`, defaults to 12):
            Number of hidden layers in the Transformer encoder.
        num_attention_heads (:obj:`int`, `optional`, defaults to 12):
            Number of attention heads for each attention layer in the Transformer encoder.
        intermediate_size (:obj:`int`, `optional`, defaults to 3072):
            Dimensionality of the "intermediate" (often named feed-forward) layer in the Transformer encoder.
        hidden_act (:obj:`str` or :obj:`Callable`, `optional`, defaults to :obj:`"gelu"`):
            The non-linear activation function (function or string) in the encoder and pooler. If string,
            :obj:`"gelu"`, :obj:`"relu"`, :obj:`"silu"` and :obj:`"gelu_new"` are supported.
        hidden_dropout_prob (:obj:`float`, `optional`, defaults to 0.1):
            The dropout probability for all fully connected layers in the embeddings, encoder, and pooler.
        attention_probs_dropout_prob (:obj:`float`, `optional`, defaults to 0.1):
            The dropout ratio for the attention probabilities.
        max_position_embeddings (:obj:`int`, `optional`, defaults to 512):
            The maximum sequence length that this model might ever be used with. Typically set this to something large
            just in case (e.g., 512 or 1024 or 2048).
        type_vocab_size (:obj:`int`, `optional`, defaults to 2):
            The vocabulary size of the :obj:`token_type_ids` passed when calling :class:`~transformers.BertModel` or
            :class:`~transformers.TFChineseBertModel`.
        initializer_range (:obj:`float`, `optional`, defaults to 0.02):
            The standard deviation of the truncated_normal_initializer for initializing all weight matrices.
        layer_norm_eps (:obj:`float`, `optional`, defaults to 1e-12):
            The epsilon used by the layer normalization layers.
        gradient_checkpointing (:obj:`bool`, `optional`, defaults to :obj:`False`):
            If True, use gradient checkpointing to save memory at the expense of slower backward pass.
        position_embedding_type (:obj:`str`, `optional`, defaults to :obj:`"absolute"`):
            Type of position embedding. Choose one of :obj:`"absolute"`, :obj:`"relative_key"`,
            :obj:`"relative_key_query"`. For positional embeddings use :obj:`"absolute"`. For more information on
            :obj:`"relative_key"`, please refer to `Self-Attention with Relative Position Representations (Shaw et al.)
            <https://arxiv.org/abs/1803.02155>`__. For more information on :obj:`"relative_key_query"`, please refer to
            `Method 4` in `Improve Transformer Models with Better Relative Position Embeddings (Huang et al.)
            <https://arxiv.org/abs/2009.13658>`__.
        use_cache (:obj:`bool`, `optional`, defaults to :obj:`True`):
            Whether or not the model should return the last key/values attentions (not used by all models). Only
            relevant if ``config.is_decoder=True``.
        glyph_embedding_dim (:obj:`int`, defaults to 1728):
            glyph_embedding_dim
        pinyin_map_len (:obj:`int`, defaults to 32):
            pinyin_map_len

    Examples::

        >>> from transformers import ChineseBertModel, ChineseBertConfig

        >>> # Initializing a BERT bert-base-uncased style configuration
        >>> configuration = ChineseBertConfig()

        >>> # Initializing a model from the bert-base-uncased style configuration
        >>> model = ChineseBertModel(configuration)

        >>> # Accessing the model configuration
        >>> configuration = model.config
    """
    model_type = "bert"

    def __init__(
        self,
        vocab_size=23236,
        hidden_size=768,
        num_hidden_layers=12,
        num_attention_heads=12,
        intermediate_size=3072,
        hidden_act="gelu",
        hidden_dropout_prob=0.1,
        attention_probs_dropout_prob=0.1,
        max_position_embeddings=512,
        type_vocab_size=2,
        initializer_range=0.02,
        layer_norm_eps=1e-12,
        pad_token_id=0,
        gradient_checkpointing=False,
        position_embedding_type="absolute",
        use_cache=True,
        glyph_embedding_dim=1728,
        pinyin_map_len=32,
        **kwargs
    ):
        super().__init__(pad_token_id=pad_token_id, **kwargs)

        self.vocab_size = vocab_size
        self.hidden_size = hidden_size
        self.num_hidden_layers = num_hidden_layers
        self.num_attention_heads = num_attention_heads
        self.hidden_act = hidden_act
        self.intermediate_size = intermediate_size
        self.hidden_dropout_prob = hidden_dropout_prob
        self.attention_probs_dropout_prob = attention_probs_dropout_prob
        self.max_position_embeddings = max_position_embeddings
        self.type_vocab_size = type_vocab_size
        self.initializer_range = initializer_range
        self.layer_norm_eps = layer_norm_eps
        self.gradient_checkpointing = gradient_checkpointing
        self.position_embedding_type = position_embedding_type
        self.use_cache = use_cache
        self.glyph_embedding_dim = glyph_embedding_dim
        self.pinyin_map_len = pinyin_map_len



# -*- coding: utf-8 -*-
from pypinyin import Style, pinyin
from transformers.models.bert import BertTokenizerFast


class ChineseBertTokenizerFast(BertTokenizerFast):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.pinyin_dict = kwargs.get("pinyin_map")
        self.id2pinyin = kwargs.get("id2pinyin")
        self.pinyin2tensor = kwargs.get("pinyin2tensor")

    def __call__(
        self,
        text,
        text_pair=None,
        add_special_tokens=True,
        padding=False,
        truncation=False,
        max_length=None,
        stride=0,
        is_split_into_words=False,
        pad_to_multiple_of=None,
        return_tensors=None,
        return_token_type_ids=None,
        return_attention_mask=None,
        return_overflowing_tokens=False,
        return_special_tokens_mask=False,
        return_offsets_mapping=False,
        return_length=False,
        verbose=True,
        **kwargs
    ):
        if isinstance(text, str):
            single = True
        else:
            single = False

        tokenizer_outputs = super().__call__(
            text=text,
            text_pair=text_pair,
            add_special_tokens=add_special_tokens,
            padding=padding,
            truncation=truncation,
            max_length=max_length,
            stride=stride,
            is_split_into_words=is_split_into_words,
            pad_to_multiple_of=pad_to_multiple_of,
            return_tensors=None,
            return_token_type_ids=return_token_type_ids,
            return_attention_mask=return_attention_mask,
            return_overflowing_tokens=return_overflowing_tokens,
            return_special_tokens_mask=return_special_tokens_mask,
            return_offsets_mapping=True,
            return_length=return_length,
            verbose=verbose,
            **kwargs
        )
        tokenizer_outputs["pinyin_ids"] = self.convert_text_to_pinyin_ids(
            text, tokenizer_outputs.offset_mapping, single
        )

        tokenizer_outputs.convert_to_tensors(
            tensor_type=return_tensors,
            prepend_batch_axis=return_tensors is not None and single,
        )

        if not return_offsets_mapping:
            del tokenizer_outputs["offset_mapping"]

        return tokenizer_outputs

    def convert_text_to_pinyin_ids(self, text, offset_mapping, single):
        # get pinyin of a sentence
        if single:
            text = [text]
            offset_mapping = [offset_mapping]

        batch_pinyin_locs = []
        for each_text in text:
            pinyin_list = pinyin(
                each_text,
                style=Style.TONE3,
                heteronym=True,
                errors=lambda x: [["not chinese"] for _ in x],
            )
            pinyin_locs = {}
            # get pinyin of each location
            for index, item in enumerate(pinyin_list):
                pinyin_string = item[0]
                # not a Chinese character, pass
                if pinyin_string == "not chinese":
                    continue
                if pinyin_string in self.pinyin2tensor:
                    pinyin_locs[index] = self.pinyin2tensor[pinyin_string]
                else:
                    ids = [0] * 8
                    for i, p in enumerate(pinyin_string):
                        if p not in self.pinyin_dict["char2idx"]:
                            ids = [0] * 8
                            break
                        ids[i] = self.pinyin_dict["char2idx"][p]
                    pinyin_locs[index] = ids
            batch_pinyin_locs.append(pinyin_locs)

        batch_pinyin_ids = []

        for offset_map, pinyin_locs in zip(offset_mapping, batch_pinyin_locs):
            pinyin_ids = []
            for offset in offset_map:
                if offset[0] in pinyin_locs and offset[1] - offset[0] == 1:
                    pinyin_ids.extend(pinyin_locs[offset[0]])
                else:
                    pinyin_ids.extend([0] * 8)

            batch_pinyin_ids.append(pinyin_ids)

        return batch_pinyin_ids[0] if single else batch_pinyin_ids



from setuptools import find_packages, setup

setup(
    name="chinesebert",
    package_dir={"": "src"},
    packages=find_packages("src"),
    version="0.0.3",
    license="MIT",
    description="ChineseBert_pytorch",
    author="Jun Yu",
    author_email="573009727@qq.com",
    url="https://github.com/JunnYu/ChineseBert_pytorch",
    keywords=["ChineseBert", "pytorch"],
    install_requires=["transformers>=4.7.0", "pypinyin", "fastcore"],
)



from dataclasses import dataclass
from typing import Optional, Union

from transformers.file_utils import PaddingStrategy
from transformers.tokenization_utils_base import PreTrainedTokenizerBase

from chinesebert.jy_utils import *


@dataclass
class DataCollatorForChineseBERT:
    tokenizer: PreTrainedTokenizerBase
    padding: Union[bool, str, PaddingStrategy] = True
    max_length: Optional[int] = None
    pad_to_multiple_of: Optional[int] = None

    def __call__(self, features):
        batch = self.tokenizer.pad(
            features,
            padding=self.padding,
            max_length=self.max_length,
            pad_to_multiple_of=self.pad_to_multiple_of,
            return_tensors="pt",
            model_input_names=["input_ids", "token_type_ids", "attention_mask"],
        )
        # batch_pinyin_ids
        batch_pinyin_ids = self.tokenizer.pad(
            features,
            padding="max_length",
            max_length=8 * batch.input_ids.size(1),
            return_tensors="pt",
            model_input_names=["pinyin_ids"],
        )
        batch["pinyin_ids"] = batch_pinyin_ids["pinyin_ids"]
        del batch_pinyin_ids

        labels = L(features).attrgot("labels").items
        if labels[0] is None:
            return batch

        batch["labels"] = torch.stack(labels, dim=0)
        return batch
